#!/usr/bin/env python3

import argparse
from os import listdir
from os.path import isfile, join
import sys
import json

from parliament import analyze_policy_string, enhance_finding


def print_finding(
    finding, minimal_output=False, json_output=False, minimum_severity="LOW"
):
    minimum_severity = minimum_severity.upper()
    finding = enhance_finding(finding)
    severity_choices = ["MUTE", "INFO", "LOW", "MEDIUM", "HIGH", "CRITICAL"]
    if severity_choices.index(finding.severity) < severity_choices.index(
        minimum_severity
    ):
        return

    if minimal_output:
        print("{}".format(finding.issue))
    elif json:
        print(
            json.dumps(
                {
                    "issue": finding.issue,
                    "title": finding.title,
                    "severity": finding.severity,
                    "description": finding.description,
                    "detail": finding.detail,
                    "location": finding.location,
                }
            )
        )
    else:
        print(
            "{} - {} - {} - {}".format(
                finding.severity, finding.title, finding.detail, finding.location
            )
        )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--aws-managed-policies",
        help="This is used with the policies directory of https://github.com/SummitRoute/aws_managed_policies",
        type=str,
    )
    parser.add_argument(
        "--auth-details-file",
        help='Provide the path to a file returned by "aws iam get-account-authorization-details"',
        type=str,
    )
    parser.add_argument(
        "--string",
        help='Provide a string such as \'{"Version": "2012-10-17","Statement": {"Effect": "Allow","Action": ["s3:GetObject", "s3:PutBucketPolicy"],"Resource": ["arn:aws:s3:::bucket1", "arn:aws:s3:::bucket2/*"]}}\'',
        type=str,
    )
    parser.add_argument("--file", help="Provide a policy in a file", type=str)
    parser.add_argument(
        "--minimal", help="Minimal output", default=False, action="store_true"
    )
    parser.add_argument(
        "--json", help="json output", default=False, action="store_true"
    )
    parser.add_argument(
        "--minimum_severity",
        help="Minimum severity to display. Options: CRITICAL, HIGH, MEDIUM, LOW, INFO",
        default="LOW",
    )
    args = parser.parse_args()

    if args.minimal and args.json:
        raise Exception("You cannot choose both minimal and json output")

    # Change the exit status if there are errors
    exit_status = 0
    findings = []

    if args.aws_managed_policies:
        filenames = [
            f
            for f in listdir(args.aws_managed_policies)
            if isfile(join(args.aws_managed_policies, f))
        ]
        for filename in filenames:
            filepath = join(args.aws_managed_policies, filename)
            with open(filepath) as f:
                contents = f.read()
                policy_file_json = json.loads(contents)
                policy_string = json.dumps(
                    policy_file_json["PolicyVersion"]["Document"]
                )
                policy = analyze_policy_string(policy_string, filepath)
                findings.extend(policy.findings)

    elif args.auth_details_file:
        with open(args.auth_details_file) as f:
            contents = f.read()
            auth_details_json = json.loads(contents)
            for policy in auth_details_json["Policies"]:
                # Ignore AWS defined policies
                if "arn:aws:iam::aws:" not in policy["Arn"]:
                    continue

                for version in policy["PolicyVersionList"]:
                    if not version["IsDefaultVersion"]:
                        continue
                    policy = analyze_policy_string(
                        json.dumps(version["Document"]), policy["Arn"]
                    )
                    findings.extend(policy.findings)

            # Review the inline policies on Users, Roles, and Groups
            for user in auth_details_json["UserDetailList"]:
                for policy in user.get("UserPolicyList", []):
                    policy = analyze_policy_string(
                        json.dumps(version["Document"]), user["Arn"]
                    )
                    findings.extend(policy.findings)
            for role in auth_details_json["RoleDetailList"]:
                for policy in role.get("RolePolicyList", []):
                    policy = analyze_policy_string(
                        json.dumps(version["Document"]), role["Arn"]
                    )
                    findings.extend(policy.findings)
            for group in auth_details_json["GroupDetailList"]:
                for policy in group.get("GroupPolicyList", []):
                    policy = analyze_policy_string(
                        json.dumps(version["Document"]), group["Arn"]
                    )
                    findings.extend(policy.findings)
    elif args.string:
        policy = analyze_policy_string(args.string)
        findings.extend(policy.findings)
    elif args.file:
        with open(args.file) as f:
            contents = f.read()
            policy = analyze_policy_string(contents)
            for finding in policy.findings:
                findings.extend(policy.findings)
    else:
        parser.print_help()
        exit(-1)

    if len(findings) == 0:
        return

    for finding in policy.findings:
        print_finding(finding, args.minimal, args.json, args.minimum_severity)
    exit(1)


if __name__ == "__main__":
    main()



"""
This library is a linter for AWS IAM policies.
"""
__version__ = "0.3.0"

import os
import json
import yaml
import re
import fnmatch
import pkg_resources

# On initialization, load the IAM data
iam_definition_path = pkg_resources.resource_filename(__name__, "iam_definition.json")
iam_definition = json.load(open(iam_definition_path, "r"))

# And the config data
config_path = pkg_resources.resource_filename(__name__, "config.yaml")
config = yaml.safe_load(open(config_path, "r"))


def enhance_finding(finding):
    if finding.issue not in config:
        raise Exception("Uknown finding issue: {}".format(finding.issue))
    config_settings = config[finding.issue]
    finding.severity = config_settings["severity"]
    finding.title = config_settings["title"]
    finding.description = config_settings.get("description", "")
    return finding


def analyze_policy_string(policy_str, filepath=None):
    """Given a string reperesenting a policy, convert it to a Policy object with findings"""

    try:
        # TODO Need to write my own json parser so I can track line numbers. See https://stackoverflow.com/questions/7225056/python-json-decoding-library-which-can-associate-decoded-items-with-original-li
        policy_json = json.loads(policy_str)
    except ValueError as e:
        policy = Policy(None)
        policy.add_finding("MALFORMED_JSON", detail="json parsing error: {}".format(e))
        return policy

    policy = Policy(policy_json, filepath)
    policy.analyze()
    return policy


def is_arn_match(resource_type, arn_format, resource):
    """
    Match the arn_format specified in the docs, with the resource
    given in the IAM policy.  These can each be strings with globbing. For example, we
    want to match the following two strings:
    - arn:*:s3:::*/*
    - arn:aws:s3:::*personalize*

    That should return true because you could have "arn:aws:s3:::personalize/" which matches both.

    Input:
    - resource_type: Example "bucket", this is only used to identify special cases.
    - arn_format: ARN regex from the docs
    - resource: ARN regex from IAM policy

    Notes:

    This problem is known as finding the intersection of two regexes.
    There is a library for this here https://github.com/qntm/greenery but it is far too slow,
    taking over two minutes for that example before I killed the process.
    The problem can be simplified because we only care about globbed strings, not full regexes,
    but there are no Python libraries, but here is one in Go: https://github.com/Pathgather/glob-intersection

    We can some cheat because after the first sections of the arn match, meaning until the 5th colon (with some
    rules there to allow empty or asterisk sections), we only need to match the ID part.
    So the above is simplified to "*/*" and "*personalize*".

    Let's look at some examples and if these should be marked as a match:
    "*/*" and "*personalize*" -> True
    "*", "mybucket" -> True
    "mybucket", "*" -> True
    "*/*", "mybucket" -> False
    "*/*", "mybucket*" -> True
    "*mybucket", "*myotherthing" -> False
    """
    if arn_format == "*" or resource == "*":
        return True

    if "bucket" in resource_type:
        # We have to do a special case here for S3 buckets
        if "/" in resource:
            return False

    arn_parts = arn_format.split(":")
    if len(arn_parts) < 6:
        raise Exception("Unexpected format for ARN: {}".format(arn_format))
    resource_parts = resource.split(":")
    if len(resource_parts) < 6:
        raise Exception("Unexpected format for resource: {}".format(resource))
    for position in range(0, 5):
        if arn_parts[position] == "*" or arn_parts[position] == "":
            continue
        elif resource_parts[position] == "*" or resource_parts[position] == "":
            continue
        elif arn_parts[position] == resource_parts[position]:
            continue
        else:
            return False

    arn_id = "".join(arn_parts[5:])
    resource_id = "".join(resource_parts[5:])

    # At this point we might have something like:
    # log-group:* for arn_id and
    # log-group:/aws/elasticbeanstalk* for resource_id

    # Look for exact match
    # Examples:
    # "mybucket", "mybucket" -> True
    # "*", "*" -> True
    if arn_id == resource_id:
        return True

    # Some of the arn_id's contain regexes of the form "[key]" so replace those with "*"
    arn_id = re.sub(r"\[.+?\]", "*", arn_id)

    # If neither contain an asterisk they can't match
    # Example:
    # "mybucket", "mybucketotherthing" -> False
    if "*" not in arn_id and "*" not in resource_id:
        return False

    # If either is an asterisk it matches
    # Examples:
    # "*", "mybucket" -> True
    # "mybucket", "*" -> True
    if arn_id == "*" or resource_id == "*":
        return True

    # We already checked if they are equal, so we know both aren't "", but if one is, and the other is not,
    # and the other is not "*" (which we just checked), then these do not match
    # Examples:
    # "", "mybucket" -> False
    if arn_id == "" or resource_id == "":
        return False

    # If one begins with an asterisk and the other ends with one, it should match
    # Examples:
    # "*/*" and "*personalize*" -> True
    if (arn_id[0] == "*" and resource_id[-1] == "*") or (
        arn_id[-1] == "*" and resource_id[0] == "*"
    ):
        return True

    # At this point, we are trying to check the following
    # "*/*", "mybucket" -> False
    # "*/*", "mybucket/abc" -> True
    # "mybucket*", "mybucketotherthing" -> True
    # "*mybucket", "*myotherthing" -> False

    # We are going to cheat and miss some possible situations, because writing something
    # to do this correctly by generating a state machine seems much harder.

    # Check situation where it begins and ends with asterisks, such as "*/*"
    if arn_id[0] == "*" and arn_id[-1] == "*":
        if arn_id[1:-1] in resource_id:
            return True
    if resource_id[0] == "*" and resource_id[-1] == "*":
        if resource_id[1:-1] in arn_id:
            return True

    # Check where one ends with an asterisk
    if arn_id[-1] == "*":
        if resource_id[: len(arn_id) - 1] == arn_id[:-1]:
            return True
    if resource_id[-1] == "*":
        if arn_id[: len(resource_id) - 1] == resource_id[:-1]:
            return True

    return False


def expand_action(action, raise_exceptions=True):
    """
    Converts "iam:*List*" to
    [
      {'service':'iam', 'action': 'ListAccessKeys'},
      {'service':'iam', 'action': 'ListUsers'}, ...
    ]
    """
    parts = action.split(":")
    if len(parts) != 2:
        raise ValueError("Action should be in form service:action")
    prefix = parts[0]
    unexpanded_action = parts[1]

    actions = []
    service_match = None
    for service in iam_definition:
        if service["prefix"] == prefix.lower():
            service_match = service

            for privilege in service["privileges"]:
                if fnmatch.fnmatchcase(
                    privilege["privilege"].lower(), unexpanded_action.lower()
                ):
                    actions.append(
                        {
                            "service": service_match["prefix"],
                            "action": privilege["privilege"],
                        }
                    )

    if not service_match and raise_exceptions:
        raise ValueError("Unknown prefix {}".format(prefix))

    if len(actions) == 0 and raise_exceptions:
        raise ValueError("Unknown action {}:{}".format(prefix, unexpanded_action))

    return actions


def get_resource_type_matches_from_arn(arn):
    """ Given an ARN such as "arn:aws:s3:::mybucket", find resource types that match it.
        This would return:
        [
            "resource": {
                "arn": "arn:${Partition}:s3:::${BucketName}",
                "condition_keys": [],
                "resource": "bucket"
            },
            "service": {
                "service_name": "Amazon S3",
                "privileges": [...]
                ...
            }
        ]
    """
    matches = []
    for service in iam_definition:
        for resource in service["resources"]:
            arn_format = re.sub(r"\$\{.*?\}", "*", resource["arn"])
            if is_arn_match(resource["resource"], arn, arn_format):
                matches.append({"resource": resource, "service": service})
    return matches


def get_privilege_matches_for_resource_type(resource_type_matches):
    """ Given the response from get_resource_type_matches_from_arn(...), this will identify the relevant privileges.
    """
    privilege_matches = []
    for match in resource_type_matches:
        for privilege in match["service"]["privileges"]:
            for resource_type_dict in privilege["resource_types"]:
                resource_type = resource_type_dict["resource_type"].replace("*", "")
                if resource_type == match["resource"]["resource"]:
                    privilege_matches.append(
                        {
                            "privilege_prefix": match["service"]["prefix"],
                            "privilege_name": privilege["privilege"],
                            "resource_type": resource_type,
                        }
                    )

    return privilege_matches


# Import moved here to deal with cyclic dependency
from .policy import Policy



import yaml


class Finding:
    """ Class for storing findings """

    issue = ""
    detail = ""
    location = {}
    severity = ""
    title = ""
    description = ""

    def __init__(self, issue, detail, location):
        self.issue = issue
        self.detail = detail
        self.location = location

    def __repr__(self):
        """ Return a string for printing """
        return "{} - {} - {}".format(self.issue, self.detail, self.location)



"""Setup script for parliament"""
import os
import re

from setuptools import find_packages, setup


HERE = os.path.dirname(__file__)
VERSION_RE = re.compile(r'''__version__ = ['"]([0-9.]+)['"]''')
TESTS_REQUIRE = [
    'coverage',
    'nose'
]


def get_version():
    init = open(os.path.join(HERE, 'parliament', '__init__.py')).read()
    return VERSION_RE.search(init).group(1)


def get_description():
    return open(os.path.join(os.path.abspath(HERE), 'README.md'), encoding='utf-8').read()


setup(
    name='parliament',
    version=get_version(),
    author='Duo Security',
    author_email='scott@summitroute.com',
    description=(
        'parliament audits your AWS IAM policies'
    ),
    long_description=get_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/duo-labs/parliament',
    entry_points={
        'console_scripts': 'parliament=parliament.cli:main'
    },
    test_suite='tests/unit',
    tests_require=TESTS_REQUIRE,
    extras_require={
        'dev': TESTS_REQUIRE + ['autoflake', 'autopep8', 'pylint']
    },
    install_requires=[
        'boto3',
        'jmespath',
        'pyyaml'
    ],
    setup_requires=['nose'],
    packages=find_packages(exclude=['tests*']),
    package_data={'parliament': ['iam_definition.json', "config.yaml"]},
    zip_safe=True,
    license='BSD 3',
    keywords='aws parliament iam lint audit',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Development Status :: 5 - Production/Stable'
    ]
)